/*
 *
 * nucleo/gl/texture/glTextureTile.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/gl/texture/glTexture.H>

#include <nucleo/gl/glUtils.H>

#ifdef __APPLE__
#include <OpenGL/glext.h>
#else
#include <GL/glext.h>
#endif

#include <stdexcept>

namespace nucleo {

  int glTextureTile::debugLevel = 0 ;
  int glTextureTile::can_generate_mipmaps = -1 ;

  // ------------------------------------------------------------------------

  static const char *
  getTargetName(GLenum target) {
    switch(target) {
    case GL_TEXTURE_2D: return "GL_TEXTURE_2D" ;
#ifdef GL_TEXTURE_RECTANGLE_EXT
    case GL_TEXTURE_RECTANGLE_EXT: return "GL_TEXTURE_RECTANGLE_EXT" ;
#endif
#ifdef GL_TEXTURE_RECTANGLE_NV
    case GL_TEXTURE_RECTANGLE_NV: return "GL_TEXTURE_RECTANGLE_NV" ;
#endif
    default: return "<unknown>" ;
    }
  }

  // ------------------------------------------------------------------------

  bool
  glTextureTile::isStandard(Image *img) {
    if (debugLevel>1)
	 std::cerr << "glTextureTile(" << this << "): trying as GL_TEXTURE_2D (" << tWidth << "x" << tHeight << ")... " << std::flush ;

    tTarget = GL_TEXTURE_2D ;
    
    // If I understand well, the GL_ARB_texture_non_power_of_two
    // extension will make this work with NPOT textures...
    // http://oss.sgi.com/projects/ogl-sample/registry/ARB/texture_non_power_of_two.txt

    glTexImage2D(GL_PROXY_TEXTURE_2D, 0,
			  tInternalFormat, tWidth, tHeight, 0, tFormat, tType, img->getData()) ;

    GLint rw=0, rh=0 ;
    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &rw) ;
    glGetTexLevelParameteriv(GL_PROXY_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &rh) ;
    bool result = ((unsigned int)rw == tWidth) && ((unsigned int)rh == tHeight) ;

    if (debugLevel>1) std::cerr << (result?"success":"failed") << std::endl ;
    return result ;
  }

  bool
  glTextureTile::worksWithExtension(Image *img) {
    GLenum proxytarget=0 ;
#if defined(GL_EXT_texture_rectangle)
    /*
	 From http://developer.apple.com/opengl/extensions/ext_texture_rectangle.html

	 Mipmap filtering is not permitted.  Since this is the case the
      default minification filter for TEXTURE_RECTANGLE_EXT targets is
      GL_LINEAR.

	 Only the GL_CLAMP, GL_CLAMP_TO_EDGE, and CLAMP_TO_BORDER wrap
      modes are allowed.  CLAMP_TO_EDGE is the default state.  REPEAT
      and MIRRORED_REPEAT_ARB are not supported with the
      TEXTURE_RECTANGLE_EXT texture target.

	 Borders are not supported.

	 Texture coordinates are addressed without being normalized from
      [0..1], instead [0..w] and [0..h] are used, where w and h are
      width and height of the texture respectively.
    */

    if (glExtensionIsSupported("GL_EXT_texture_rectangle")) {
	 tTarget = GL_TEXTURE_RECTANGLE_EXT ;
	 proxytarget = GL_PROXY_TEXTURE_RECTANGLE_EXT ;
    } else {
	 if (debugLevel>1)
	   std::cerr  << "glTextureTile(" << this << "): GL_EXT_texture_rectangle is known but not supported" << std::endl ;
    }
#elif defined(GL_NV_texture_rectangle)
    /*
	 From http://oss.sgi.com/projects/ogl-sample/registry/NV/texture_rectangle.txt

	 (...) non-power-of-two dimensioned (NPOTD) textures have
      limitations that do not apply to power-of-two dimensioned (POT)
      textures.  NPOTD textures may not use mipmap filtering; (...)
      NPOTD textures support only the GL_CLAMP, GL_CLAMP_TO_EDGE, and
      GL_CLAMP_TO_BORDER_ARB wrap modes; (...).  NPOTD textures do not
      support an optional 1-texel border; (...).

      NPOTD textures are accessed by non-normalized texture
      coordinates.  So instead of thinking of the texture image lying
      in a [0..1]x[0..1] range, the NPOTD texture image lies in a
      [0..w]x[0..h] range.

      The default wrap state for a texture rectangle is
      GL_CLAMP_TO_EDGE.  The normal default wrap state is GL_REPEAT,
      but that mode is not allowed for rectangular textures.

      The default minification filter for a texture rectangle is
      GL_LINEAR.  The normal default minification filter state is
      GL_NEAREST_MIPMAP_LINEAR, but that mode is not allowed for
      rectangular textures because mipmapping is not supported.
    */

    if (glExtensionIsSupported("GL_NV_texture_rectangle")) {
	 tTarget = GL_TEXTURE_RECTANGLE_NV ;
	 proxytarget = GL_PROXY_TEXTURE_RECTANGLE_NV ;
    } else {
	 if (debugLevel>1)
	   std::cerr  << "glTextureTile(" << this << "): GL_NV_texture_rectangle is known but not supported" << std::endl ;
    }
#endif

    if (!proxytarget) {
	 if (debugLevel>1) std::cerr  << "glTextureTile(" << this << "): no texture_rectangle extension available" << std::endl ;
	 return false ;
    }

    if (debugLevel>1)
	 std::cerr << "glTextureTile(" << this << "): trying as " << getTargetName(tTarget) << " (" << tWidth << "x" << tHeight << ")... " << std::flush ;

    while (glGetError()!=GL_NO_ERROR) ; // we might have generated some errors
    glTexImage2D(proxytarget, 0,
			  tInternalFormat, tWidth, tHeight, 0, tFormat, tType, img->getData()) ;

    GLint rw=0, rh=0 ;
    glGetTexLevelParameteriv(proxytarget, 0, GL_TEXTURE_WIDTH, &rw) ;
    glGetTexLevelParameteriv(proxytarget, 0, GL_TEXTURE_HEIGHT, &rh) ;
    bool result = ((unsigned int)rw == tWidth) && ((unsigned int)rh == tHeight) ;

    if (debugLevel>1) std::cerr << (result?"success":"failed") << std::endl ;
    return result ;
  }

  // ------------------------------------------------------------------------

  glTextureTile::glTextureTile(Image *img,
						 unsigned int xp, unsigned int yp, unsigned int w, unsigned int h,
						 glTexture *m) {
    if (can_generate_mipmaps==-1)
	 can_generate_mipmaps= glExtensionIsSupported("GL_SGIS_generate_mipmap") ;

    master = m ;
    x = xp ; y = yp ;
    tWidth = width = w ; tHeight = height = h ;

    glGenTextures(1, &texture) ;
    if (!glImageEncodingParameters(img->getEncoding(),
							&tFormat, &tInternalFormat, &tAlignment, &tType)) {
	 std::cerr << "glTextureTile: glTexture should have checked the image encoding. How did we get there?" << std::endl ;
	 return ;
    }

    memoryholder.linkDataFrom(*img) ;

    bool foundTarget = false ;

    switch (master->trePolicy) {
    case FIRST_CHOICE:
	 foundTarget = worksWithExtension(&memoryholder) || isStandard(&memoryholder) ;
	 break ;
    case SECOND_CHOICE:
	 foundTarget = isStandard(&memoryholder) || worksWithExtension(&memoryholder) ;
	 break ;
    case DONT_USE:
	 foundTarget = isStandard(&memoryholder) ;
	 break ;
    } 

    if (foundTarget) {

	 setup(&memoryholder, false) ;

    } else {

	 tWidth=1 ; while (tWidth<width) tWidth*=2 ;
	 tHeight=1 ; while (tHeight<height) tHeight*=2 ;

	 // Allocate a POT memory buffer and copy the tile data in it
	 Image::Encoding encoding = memoryholder.getEncoding() ;
	 memoryholder.prepareFor(tWidth,tHeight,encoding) ;
	 unsigned int bpp = memoryholder.getBytesPerPixel() ;
	 unsigned int rowLength = bpp*width ;
	 unsigned int srcRowLength = bpp*img->getWidth() ;
	 unsigned int dstRowLength = bpp*memoryholder.getWidth() ;
	 unsigned char *src = img->getData() + srcRowLength*y + bpp*x ;
	 unsigned char *dst = memoryholder.getData() ;
	 for (unsigned int r=0; r<height; ++r) {
	   memcpy(dst, src, rowLength) ;
	   src += srcRowLength ;
	   dst += dstRowLength ;
	 }

	 foundTarget = isStandard(&memoryholder) ;

	 if (foundTarget) {
	   x = y = 0 ;
	   setup(&memoryholder, false) ;
	   x = xp ; y = yp ;
	 }

    }

    while (glGetError()!=GL_NO_ERROR) ; // we might have generated some errors
    if (!foundTarget) throw std::runtime_error("glTextureTile: can't create texture") ;

    if (debugLevel) {
	 std::cerr << "glTextureTile(" << this << "): " ;
	 std::cerr << getTargetName(tTarget) << " " ;
	 std::cerr << width << "x" << height << " " ;
	 if (tTarget==GL_TEXTURE_2D) std::cerr << "(" << tWidth << "x" << tHeight << ") " ;
	 std::cerr << x << "," << y << std::endl ;
    }
  }

  void
  glTextureTile::update(Image *img) {
    setup(img, true) ;
  }

  void
  glTextureTile::setup(Image *img, bool update) {
    glBindTexture(tTarget, texture) ;
    
    glPixelStorei(GL_UNPACK_ALIGNMENT, tAlignment) ;
    glPixelStorei(GL_UNPACK_ROW_LENGTH, img->getWidth()) ;
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, x) ;
    glPixelStorei(GL_UNPACK_SKIP_ROWS, y) ;

    if (update) {
	 glTexSubImage2D(tTarget, 0,
				  0, 0, width, height,
				  tFormat, tType, img->getData()) ;
    } else {
#ifdef __APPLE__
	 // For the details of these magic tricks, see
	 // http://developer.apple.com/samplecode/Sample_Code/Graphics_3D/TextureRange.htm

	 glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, master->useClientStorage ? 1 : 0) ;

	 glTextureRangeAPPLE(tTarget, img->getSize(), img->getData()) ;

	 // AGP texturing please
	 glTexParameterf(tTarget, GL_TEXTURE_PRIORITY, 0.0f) ;

	 // Use *CACHED* for VRAM, *SHARED* for AGP and *PRIVATE* for
	 // the normal texturing path
	 glTexParameteri(tTarget, GL_TEXTURE_STORAGE_HINT_APPLE , GL_STORAGE_SHARED_APPLE) ;
#endif

	 glTexParameteri(tTarget, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE) ;
	 glTexParameteri(tTarget, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE) ;

	 if (tTarget==GL_TEXTURE_2D) {
#if GL_SGIS_generate_mipmap
	   if (can_generate_mipmaps && master->generateMipmaps) {
		if (debugLevel>1) std::cerr << "glTextureTile(" << this << "): mipmaps will be automatically generated" << std::endl ;
		glTexParameteri(tTarget, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR) ;
		glTexParameteri(tTarget, GL_GENERATE_MIPMAP_SGIS, GL_TRUE) ;
	   } else
#endif
		glTexParameteri(tTarget, GL_TEXTURE_MIN_FILTER, GL_NEAREST) ;
	   glTexParameteri(tTarget, GL_TEXTURE_MAG_FILTER, GL_NEAREST) ;

	 } else {
	   glTexParameteri(tTarget, GL_TEXTURE_MAG_FILTER, master->magFilter) ;
	   glTexParameteri(tTarget, GL_TEXTURE_MIN_FILTER, master->minFilter) ;
	 }

	 glTexImage2D(tTarget, 0, tInternalFormat,
			    tWidth, tHeight, 0,
			    tFormat, tType, img->getData()) ;
    }

#if 0
    glPixelStorei(GL_UNPACK_ROW_LENGTH, 0) ;
    glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0) ;
    glPixelStorei(GL_UNPACK_SKIP_ROWS, 0) ;
#endif
  }

  glTextureTile::~glTextureTile(void) {
    // glFinish() ;
    glDeleteTextures(1, &texture) ;
  }

  // ------------------------------------------------------------------------

}
